from supervisely.nn.benchmark.comparison.detection_visualization.vis_metrics.avg_precision_by_class import (
    AveragePrecisionByClass,
)
from supervisely.nn.benchmark.comparison.detection_visualization.vis_metrics.calibration_score import (
    CalibrationScore,
)
from supervisely.nn.benchmark.comparison.detection_visualization.vis_metrics.explore_predicttions import (
    ExplorePredictions,
)
from supervisely.nn.benchmark.comparison.detection_visualization.vis_metrics.localization_accuracy import (
    LocalizationAccuracyIoU,
)
from supervisely.nn.benchmark.comparison.detection_visualization.vis_metrics.outcome_counts import (
    OutcomeCounts,
)
from supervisely.nn.benchmark.comparison.detection_visualization.vis_metrics.overview import (
    Overview,
)
from supervisely.nn.benchmark.comparison.detection_visualization.vis_metrics.pr_curve import (
    PrCurve,
)
from supervisely.nn.benchmark.comparison.detection_visualization.vis_metrics.precision_recal_f1 import (
    PrecisionRecallF1,
)
from supervisely.nn.benchmark.comparison.detection_visualization.vis_metrics.speedtest import (
    Speedtest,
)
