from supervisely.nn.benchmark.evaluation.base_evaluator import BaseEvaluator
from supervisely.nn.benchmark.evaluation.object_detection_evaluator import ObjectDetectionEvaluator
from supervisely.nn.benchmark.evaluation.instance_segmentation_evaluator import InstanceSegmentationEvaluator
